# Dependency Injection

As our programs grow in size, we often find that different parts of the app need to communicate with other modules. When module A requires module B to run, we say that B is a dependency of A.

One of the most common ways to get access to dependencies is to simply import a file.

```ts
// in A.ts
import { B } from 'B'; // a dependency!
B.foo(); // using B
```

In many cases, simply import other code is sufficient. However there are times where we need to provide dependencies in a more sophisticated way. For instance:

* What if we wanted to substitute out the implementation of B for MockB during testing?
* What if we wanted to share a single instance of the B class across our whole app (e.g. the Singleton pattern)
* What if we wanted to create a new instance of the B class every time it was used? (e.g. the Factory pattern)

Dependency Injection can solve these problems. Dependency Injection (DI) is a system to make parts of our program accessible to other parts of the program - and we can configure how that happens.

> One way to think about an injector is as a replacement for the `new` operator

The term Dependency Injection is used to describe both a design pattern (that used in many different frameworks) and also the specific implementation DI library that is built-in to Angular.

The major benefit of using dependency injection is that the client component doesn’t have to be aware of how to create the dependencies, all the component needs to know is how to interact with those dependencies.

## Injections Example: PriceService

通过示例展示了使用依赖注入带来的诸多好处

## Don't call us, we'll call you

在以前 Component 大包独揽，组件对其依赖项知道得太多了，这种紧耦合使得组件变得异常脆弱。

而使用依赖注入，只要依赖项的对外接口没有修改，那么所作的其他修改对依赖其运行的组件没有丝毫影响。

![](/resource/images/ng2/di.png)

Traditionally, if a component A depends on component B, what would happen is that an instance of B would be created inside A. This implies that now A depends on B.

Angular uses the Dependency Injection to change things around in a way that if we need component B inside component A, we expect that B will be passed to A.

This brings many advantages over the traditional scenario. One example of an advantage is that, if we’re testing A in isolation we can easily create a mocked version of B and inject it into A.

## Dependency Injection Parts

To register a dependency we have to bind it to something that will identify that dependency. This identification is called the dependency **token**. For instance, if we want to register the URL of an API, we can use the string API_URL as the token. Similarly, if we’re registering a class, we can use the class itself as its token as we’ll see below.

Dependency injection in Angular has three pieces:

* the **Provider** (also often referred to as a binding) maps a token (that can be a string or a class) to a list of dependencies. It tells Angular how to create an object, given a token.
* the **Injector** that holds a set of bindings and is responsible for resolving dependencies and injecting them when creating objects
* the **Dependency** that is what’s being injected
 
We can think of the role of each piece as illustrated below:

![](/resource/images/ng2/di2.png)

There are a lot of different options when dealing with DI, so let’s see how each of them work.

One of the most common cases is providing a service or value that is the same across our whole application. This scenario would be what we would use 99% of the time in our apps.

## Playing with an Injector

```ts
import { Component, ReflectiveInjector } from '@angular/core';

// The injectable service
class MyService {
  getValue(): string { return 'a value'; }
}

@Component({
  selector: 'di-sample-app',
  template: `<button (click)="invokeService()">Get Value</button>`
})
class DiSampleApp {
  myService: MyService;

  constructor() {
    let injector: any = ReflectiveInjector.resolveAndCreate([MyService]);
      // we are using a static method from ReflectiveInjector called resolveAndCreate.
      // That method is responsible for creating a new injector. 
      // The parameter we pass in is an array with all the injectable things we want this new injector to know
    this.myService = injector.get(MyService);
    console.log('Same instance?', this.myService === injector.get(MyService));  // true
  }

  invokeService(): void { console.log('MyService returned', this.myService.getValue()); }
}
```

> The ReflectiveInjector is a concrete implementation of Injector that uses reflection to look up the proper parameter types. While there are other injectors that are possible ReflectiveInjector is the “normal” injector we’ll be using in most apps.

Notice that, since we’re using our own injector, we didn’t have to add MyService to the NgModule providers list.

## Providing Dependencies with `NgModule`

However, normally we do need to tell our NgModule about the providers of things we will inject.

For instance, say we wanted to allow a singleton instance MyService to be injected across our app.

In order to be able to inject these things we must add them to the providers key of a NgModule.

```ts
@NgModule({
  declarations: [
    MyAppComponent,
    // other components ...
  ],
  providers: [ MyService ] // <--- here
  })
  class MyAppModule {  }
```

Now MyAppComponent could inject MyService in the constructor like so:

```ts
export class MyAppComponent {
  constructor(private myService: MyService /* <-- injected */) {
    // do something with myService here
  }

  // ...
}
```

When we put the class itself into the list of providers like this:

```ts
providers: [ MyService ]
```

That is telling Angular that we want to provide a singleton instance of MyService whenever MyService is injected. Because this pattern is so common, the class by itself is shorthand notation for the following, equivalent configuration:

```ts
providers: [
  { provide: MyComponent, useClass: MyComponent }
]
```

There are many different ways of injecting things beyond creating an instance of a class. let’s take a look.

## Providers

One of the neat things about Angular’s DI system is that there are several ways we can configure
the injection. For instance we can:

* Inject a (singleton) instance of a class
* Call any function and inject the return value of that function
* Inject a value
* Create an alias

### Using a Class

Injecting a singleton instance of a class is probably the most common type of injection. Here’s how we configure it:

```ts
{ provide: MyComponent, useClass: MyComponent }
```

What’s interesting to note is that the provide configuration method takes two keys. The first provide is the token that we use to identify the injection and the second useClass is how and what to inject.

So here we’re mapping the MyComponent class to the MyComponent token. In this case, the name of the class and the token match. This is the common case, but know that the token and the injected thing don’t have to have the same name.

As we’ve seen above, in this case the injector will create a singleton behind the scenes and return the same instance every time we inject it.
Of course, the first time it is injected, it hasn’t been instantiated yet, so when creating the MyComponent instance for the first time, the DI system will trigger the class constructor method.

### Using a Factory

If we need to pass in parameters when creating a service, we would need to use a factory instead.

```ts
class ParamService {
  constructor(private phrase: string) {
    console.log('ParamService is being created with phrase', phrase);
  }
  getValue(): string { return this.phrase; }
}
```

When we use a factory injection, we write a function that can return any object.

```ts
{
  provide: MyComponent,
  useFactory: () => {
    if (loggedIn) { return new MyLoggedComponent(); }
    return new MyComponent();
  }
}
```

Factories can also have dependencies:

```ts
{
  provide: MyComponent,
  useFactory: (user) => {
    if (user.loggedIn()) { return new MyLoggedComponent(user); }
    return new MyComponent();
  },
  deps: [ User ]
}
```

So if we wanted to use our ParamService from above, we can wrap it with useFactory like so:

```ts
@NgModule({
  declarations: [ DiSampleApp ],
  imports: [ BrowserModule ],
  bootstrap: [ DiSampleApp ],
  providers: [
    SimpleService,  // 多数情况下都是这样采用简写，
                    // 相当于 { provide: SimpleService, useClass: SimpleService }
    { provide: ParamService, useFactory: (): ParamService => new ParamService('YOLO') }
  ]
})
class DiSampleAppAppModule {  }
```

Using a factory is the most powerful way to create injectables, because we can do whatever we want within the factory function.

### Using a Value

This is useful when we want to register a constant that can be redefined by another part of the application or even by environment (e.g. test or production).

```ts
{ provide: 'API_URL', useValue: 'http://my.api.com/v1' }
```

### Using an alias

We can also make an alias to reference a previously registered token, like so:

```ts
{ provide: NewComponent, useExisting: MyComponent }
```

## Dependency Injection in Apps

When writing our apps there are three steps we need to take in order to perform an injection:

1. Create the service class
2. Declare the dependencies on the receiving component and
3. Configure the injection (i.e. register the injection with Angular in our NgModule )

The first thing we do is create the service class, that is, the class that exposes some behavior we want to use. This will be called the injectable because it is the thing that our components will receive via the injection.

Here is how we would create a service:

```ts
export class ApiService {
  get(): void { console.log('Getting resource...'); }
}
```

Now that we have the thing to be injected, we have to take the next step, which is declare the
dependencies we want to receive when Angular creates our component.
Earlier we used the Injector class directly, but Angular provides two shortcuts for us we can use
when writing our components.
The first and typical way of doing it, is by declaring the injectables we want in our component’s
constructor.
To do that, we require the service:

```ts
import { ApiService } from 'services/ApiService';

@Component({ ... })
class DiSampleApp {
  constructor(private apiService: ApiService) { ... }
...
}
```

When we declare the injection in our component constructor, Angular will do some reflection to
figure out what class to inject. That is, Angular will see that we are looking for an object of type
ApiService in the constructor and check the DI system for an appropriate injection.
Sometimes we need to give Angular more hints about what we’re trying to inject. In that case we
use the second method by using the `@Inject` annotation:

```ts
class DiSampleApp {
  private apiService: ApiService;
  constructor(@Inject(ApiService) apiService) {
  this.apiService = apiService;
}
```
